/**
 * 定义一个功能页的map
 * var Pages={
 *     'home':'pages/home',
 *     'list':'pages/list',
 *     'detail':'pages/detail',
 *     'edit':'pages/edit',
 *     'report':'pages/report'
 * };
 *
 * 创建一个NavigationController元素
 * <NavigationController pages={Pages} rootPage={'home'} rootParams={'test'}></NavigationController>
 */


import React, { Component } from 'react'
import Page from '../page/Page'

/**
 * </br><span style='font-size:14px;color:#74af1f;'>页面控制器</span>：
 * <ul style='font-size:14px;color:#74af1f;'>
 * <li>1.采用栈的方式保持页面对象</li>
 * <li>2.栈的页面和顺序保持与浏览器的history同步</li>
 * <ul>
 */
class NavigationController extends Component{

    /**
     * NavigationController的构造函数
     * @param props
     * @param context
     */
    constructor(props, context) {
        super(props, context);
        /*
         * 当进行栈处理时,打开stackHash标志,避免浏览器hash改变时,触发其他的逻辑
         */
        this.stackHash=false;
        this.replaceHash=null;
        this.pushHash=null;
        this.stackRecover=false;
        /*
         * 第一次是否渲染完成的标志,当完成第一次渲染后,之后的render都会记录栈到localstorage
         */
        this.firstRender=false;
        this.popParams=null;
        /*
         * page class对象被确认为一个page对象,并且加载完成之后,加入到map
         */
        this.loadMap={};
        /*
         * 当前被激活的page
         */
        this.currentPage = null;

        this.state = {
            pageStack:[]
        };

        this._initExcuteFunction();
        this._initBind();
        this._initRestoreHash();
    }

    /**
     * 绑定native的调用方法
     * @private
     */
    _initExcuteFunction(){
        //在window上挂载接入Native的pageclick处理
        window.executePageClick = (uri, pageId) => {
            this._onExecutePageClick(uri,pageId);
        };
        //在window上挂载接入保存页面状态的方法
        window.executePageStateStore = () => {
            this._onExecutePageStateStore();
        };
    }

    /**
     * 方法绑定作用域
     * @private
     */
    _initBind(){
        this._renderPage = this._renderPage.bind(this);
        this._hashListener = this._hashListener.bind(this);
    }

    /**
     * 恢复场景下的hash提前处理
     * @private
     */
    _initRestoreHash(){
        if(window.mutants && window.mutants.isRestore === true){
            console.log("现在是android被杀掉的情况!");
            const pageState = this._restoreStackState().pageStack;
            const ilen = pageState.length;
            if(ilen>1){
                for(let i=1;i<ilen;i++){
                    //恢复history的栈历史
                    history.pushState({},'','#'+pageState[i].hash);
                }
            }
            this.stackRecover = true;
        }
    }

    render(){
        if(this.firstRender) this._saveLocalStorage(this.state);
        this._synchronizeHashOnPush();
        let pages=this.state.pageStack.map(this._renderPage);
        const maskStyle = { position:"fixed", left:"0px", top:"0px", right:"0px", bottom:"0px", background:"#FFFFFF",zIndex:"99"};
        return (<div className="navigationController">
            {pages}
            <div className="navigationPageMask" style={maskStyle}></div>
        </div>)
    }

    componentDidMount(){
        this.firstRender=true;
        //启动对hash的监听
        window.onhashchange=this._hashListener;
        //如果是android,监听返回键
        if(window.mutants && window.mutants.env && window.mutants.env.os && window.mutants.env.os === 'android'){
            document.addEventListener('backbutton', this._onAndroidBackBtnClick.bind(this), false);
        }
        if(this.stackRecover) {
            this._initStack('recover');
            this.stackRecover = false;
        }
        else{
            this._initStack(window.location.hash);
        }
    }

    componentDidUpdate(){
        const page = this._getCurrentPage();
        if(page){
            this.currentPage=this._getPageInstance(page);
            //通知当前页面，页面激活了
            this.currentPage.willActive();
            this._popWithParams(this.currentPage);
        }
    }

    shouldComponentUpdate(nextProps,nextState){
        return this._isPageClassLoaded(nextProps,nextState);
    }

    /**
     * 判断当前应用打开方式，然后进行初始化
     * 1.当hash为空时，表示新打开浏览器进入，此时加载应用定义的rootPage
     * 2.当hash不为空时，检查是否有保存在localStorage中的持久化栈数据，如果有则按数据恢复页面栈
     * @private
     */
    _initStack(hash){
        //如果当前hash为空，表示新打开浏览器进入
        if(hash===undefined || hash===""){
            //清除localStorage中的栈，然后从rootPage开始
            this._push(this.props.rootPage,this.props.rootParams,null,true);
        }
        //从localStorage中恢复
        else if(window.localStorage && window.localStorage.pageStack){
            //检查localStorage中的栈，然后恢复栈队列，到最后一页
            let pageStack=this._restoreStackState(),
                stack=pageStack.pageStack,
                pageNum=stack.length;
            //hash是否对应当前栈的最后一个页面，如果不对应，重新初始化
            if(pageNum>0 && (hash === 'recover' || hash.indexOf(stack[pageNum-1].name)>-1)){
                this.setState(pageStack);
            }
            else{
                this._push(this.props.rootPage,this.props.rootParams,null,true);
            }
        }
        //重新开始
        else{
            //没有localStorage，从rootPage开始
            this._push(this.props.rootPage,this.props.rootParams,null,true);
        }
    }

    /**
     * 当push操作后，将hash history 进行同步操作
     * @private
     */
    _synchronizeHashOnPush(){
        if(this.pushHash){
            this.stackHash=true;
            this._hash(this.pushHash);
            this.pushHash=null;
            //之前被激活的页面执行未激活
            if(this.currentPage && this.currentPage.unActive && typeof this.currentPage.unActive === 'function'){
                this.currentPage.willUnActive();
            }
        }
    }

    /**
     * 返回一个Page Element
     * @param item
     * @param index
     * @returns {XML}
     * @private
     */
    _renderPage(item,index){

        let PageView;
        let PageComponent;
        let stack=this.state.pageStack;
        let currentIndex=stack.length-1;
        let showIndex = index;

        PageView=this.loadMap[item.name];

        if(PageView==null){
            throw new Error(item.name+'Page还没有加载完成');
        }
        else if(PageView=="unRegisterPage"){
            throw new Error(item.name+'Page没有注册');
        }
        else if(PageView=="invalidPage"){
            throw new Error(item.name+'不是一个可使用的Page对象');
        }
        else if(PageView=="waitPage"){
            return (<div key={index}></div>);
        }
        //判断引入的页面对象是否来自Page的继承
        else if(this._isPageObject(PageView)){
            let className,style;
            const pageContext = {
                pageId: item.hash,
                params: item.params,
                state: item.state
            };
            if(index==currentIndex){
                className="navigationPage currentPage pageIndex"+index;
                PageComponent=(<PageView
                    {...this.props}
                    ref={item.hash}
                    pageContext={pageContext}
                    pageId={index}
                    navigationController={this}
                    params={item.params}
                    state={item.state}
                />);
                showIndex = 100;
            }
            //优化：如果是来自localStorage的整体恢复，非当前显示页不进行实例化
            else if(!item.isRestored){
                className="navigationPage hidePage pageIndex"+index;
                PageComponent=(<PageView
                    {...this.props}
                    ref={item.hash}
                    pageContext={pageContext}
                    pageId={index}
                    navigationController={this}
                    params={item.params}
                    state={item.state}
                />);
            }
            style={position:'fixed',left:'0px',right:'0px',top:'0px',bottom:'0px',zIndex:showIndex};
            return (<div key={index} className={className} style={style}>{PageComponent}</div>);
        }
        else{
            throw new Error(item.name+":这不是一个可使用的Page对象");
        }
    }

    /**
     * 获得当前页面的实例
     * @private
     */
    _getPageInstance(page){
        if(page.getWrappedInstance){
            page=page.getWrappedInstance();
            return this._getPageInstance(page);
        }
        else{
            return page;
        }
    }

    /**
     * 判断pop是否携带参数，如果携带参数，执行当前页面的pageBackReceiveResult方法
     * @private
     */
    _popWithParams(page){
        if(this.popParams){
            if(page.pageBackReceiveResult){
                page.pageBackReceiveResult(this.popParams.action,this.popParams.params,this.popParams.pageCode);
                this.popParams=null;
            }
            else{
                throw new Error('当前Page对象上无法找到pageBackReceiveResult,无法触发上一个Page带回的结果');
            }
        }
    }

    /**
     * 判断当前栈页面的Page对象是否加载
     * 如果需要异步加载，return false，中止执行render，等待加载完成后再次触发检测
     * @private
     */
    _isPageClassLoaded(nextProps,nextState){
        let i,
            pageName,
            PageModules,
            loadStatus=true,
            pageStack=nextState.pageStack,
            ilen=pageStack.length,
            self=this;

        for(i=0;i<ilen;i++){
            pageName=pageStack[i].name;
            //如果是恢复模式，则栈最后的显示页面的isRestored状态清空
            if(i==ilen-1){
                delete pageStack[i].isRestored;
            }
            //isRestored为true的页面不需要渲染
            if(pageStack[i].isRestored===true){
                this.loadMap[pageName]="waitPage";
            }
            else{
                //如果pageModues在loadMap中不存在，则需要获取
                if(!this.loadMap[pageName] || this.loadMap[pageName]=="waitPage"){
                    PageModules=this.props.pages[pageName];
                    if(PageModules==null){
                        this.loadMap[pageName]="unRegisterPage";
                    }
                    //如果有getComponet方法，则表示页面是异步加载模式
                    else if(PageModules.getComponent!=null){
                        this._asynLoadPageClass({
                            pageName:pageName,
                            PageModules:PageModules,
                            nextState:nextState
                        });
                        loadStatus=false;
                    }
                    //如果没有getComponent方法，检测页面是否直接就是Page对象
                    else if(this._isPageObject(PageModules)){
                        this.loadMap[pageName]=PageModules;
                    }
                    else{
                        this.loadMap[pageName]="invalidPage";
                    }
                }
            }
        }
        return loadStatus;
    }

    /**
     * 加载页面类
     * @param pageName
     * @param callback
     * @private
     */
    _loadPageClass(pageName,callback){
        //查看Map中是否已经load了
        if(this.loadMap[pageName]){
            callback();
        }
        else{
            let PageModules=this.props.pages[pageName];
            if(PageModules==null){
                throw new Error(pageName+'Page没有注册');
            }
            //如果有getComponet方法，则表示页面是异步加载模式
            else if(PageModules.getComponent!=null){
                this._asynLoadPageClass({
                    pageName:pageName,
                    PageModules:PageModules
                },callback);
            }
            //如果没有getComponent方法，检测页面是否直接就是Page对象
            else if(this._isPageObject(PageModules)){
                this.loadMap[pageName]=PageModules;
                callback();
            }
            else{
                throw new Error(pageName+'不是一个可使用的Page对象');
            }
        }
    }

    /**
     * 异步加载页面类，加载完成后判断返回的页面类是否一个继承自Page的有效类
     * @param params
     * @param callback
     * @private
     */
    _asynLoadPageClass(params,callback){
        let self=this;
        params.PageModules.getComponent(function(pageClass){
            if(self._isPageObject(pageClass)){
                self.loadMap[params.pageName]=pageClass;
            }
            else{
                self.loadMap[params.pageName]="invalidPage";
                //throw new Error(pageName+'不是一个可使用的Page对象');
            }
            //根据是否有callback，来进行不同的处理
            if(callback){
                callback();
            }
            else{
                self.setState(params.nextState);
            }
        });
    }

    /**
     * @description 判断页面对象是否来自Page类的继承
     * @param child
     * @returns true/false
     * @private
     */
    _isPageObject(child){
        //如果是被connect包装过的
        if(child.WrappedComponent && typeof(child.WrappedComponent)=="function"){
            child=child.WrappedComponent;
        }
        if(Object.getPrototypeOf(child)===Page){
            return true;
        }
        else if(Object.getPrototypeOf(child)!==null){
            return this._isPageObject(Object.getPrototypeOf(child));
        }
        else{
            return false;
        }
    }

    /**
     * 改变页面的hash
     * @private
     */
    _hash(hash){
        window.location.hash=hash;
    }

    /**
     * @description 监听hash的变化，如果来栈改变带来的hash改变时，不进行处理
     * @private
     */
    _hashListener(params){
        if(this.stackHash){
            //如果是来自栈的replace方式，进行hash的替换
            if(this.replaceHash!=null && this.replaceHash!=""){
                this._hash(this.replaceHash);
                this.replaceHash=null;
            }
            else{
                this.stackHash=false;
            }
        }
        //否则来自history.back或者URL输入
        else{
            this._popTo({isIgnoreHash:true});
        }
    }

    /**
     * @description 从LocalStorage里恢复栈
     * @private
     */
    _restoreStackState(){
        var i,ilen,
            stack=[],
            pageStack=window.localStorage.pageStack;
        if(pageStack!=null && pageStack!='' && pageStack!=='""' && pageStack!='null'){
            stack=JSON.parse(pageStack).pageStack;
            if(stack && stack.length){
                ilen=stack.length;
                for(i=0;i<ilen;i++){
                    stack[i].isRestored=true;
                }
            }
        }
        return {pageStack:stack};
    }

    /**
     * @description 保存数据到LocalStorage
     * @private
     */
    _saveLocalStorage(data){
        if(window.localStorage){
            const stackJsonString =  JSON.stringify(data);
            window.localStorage.pageStack = stackJsonString;
        }
        else{
            throw new Error("no localStorage");
        }
    }

    /**
     * @description 检查页面的参数，参数中不能有function
     * @private
     */
    _checkPageParams(params){
        let ptype=typeof(params);
        if(params===undefined || params===null){
            return true;
        }
        else if(ptype=='number'||ptype=='string'||ptype=='boolean'){
            return true;
        }
        else if(ptype=='function'){
            return false;
        }
        else if(params instanceof Array){
            let i,len=params.length;
            for(i=0;i<len;i++){
                if(!this._checkPageParams(params[i])) return false;
            }
            return true;
        }
        else if(params instanceof Object){
            let j;
            for(j in params){
                if(!this._checkPageParams(params[j])) return false;
            }
            return true;
        }
        else{
            return false;
        }
    }

    /**
     * @description 处理Native方式下的NavigateBar的click事件
     * @private
     */
    _onExecutePageClick(uri,pageId){
        let currentPage = this._getCurrentPage();
        if(currentPage){
            currentPage=this._getPageInstance(currentPage);
            currentPage.executeNativeClick && currentPage.executeNativeClick(uri);
        }
    }

    /**
     * @description 处理Native方式下的保存现场事件
     * @private
     */
    _onExecutePageStateStore(){

        const {pageStack} = this.state;

        let i;
        let page;
        let state;
        const ilen = this.state.pageStack.length;

        for(i=0;i<ilen;i++){
            page = this.refs[this.state.pageStack[i].hash];
            //page是否被生成,有可能没有被重新生成
            if(page){
                page = this._getPageInstance(page);
                if(page.saveInstanceState) {
                    state = page.saveInstanceState();
                    if(state === null || state === undefined){
                        state = '';
                    }
                    this.state.pageStack[i].state = state;
                }
                else{
                    this.state.pageStack[i].state = '';
                }
            }
        }
        this._saveLocalStorage(this.state);
    }

    /**
     * @description android下的物理返回键盘处理
     * @private
     */
    _onAndroidBackBtnClick() {
        let currentPage = this._getCurrentPage();
        if(currentPage){
            currentPage=this._getPageInstance(currentPage);
            currentPage.androidBackAction && currentPage.androidBackAction();
        }
    }

    /**
     * @description 获取当前页面的对象
     * @private
     */
    _getCurrentPage() {
        const {pageStack} = this.state;
        const len = pageStack.length;
        if(len>0){
            const currentPage = pageStack[len-1].hash;
            return this.refs[currentPage];
        }
        return false;
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>增加一个栈节点页面</span></br></br>
     * @param name {String} 在传入的页面配置对象Pages中的key值
     * @param pageParams {JSONObject} 带给新页面的页面参数
     * @param pageCode {String} 给新页面一个标识，用于新页面获得结果后返回前一页面时，对新页面的来源判断
     */
    push(name,pageParams,pageCode){
        var self=this;
        this._loadPageClass(name,function(){
            self._push(name,pageParams,pageCode,false);
        });
    }

    /**
     * @description 给页面栈增加一个页面节点
     * @private
     */
    _push(name,pageParams,pageCode,isIgnoreHash){
        let pageData;
        let stack = this.state.pageStack;
        let hash;
        if(this.props.pages[name]){
            const pageId = (new Date()).getTime();
            hash = name + '/' + pageId;
            pageData={
                pageId:pageId,
                name:name,
                params:pageParams||'',
                hash:hash
            };
            if(pageCode!=null) pageData.pageCode=pageCode;
            stack.push(pageData);
            if(!isIgnoreHash){
                this.pushHash = hash;
            }
            this.setState({pageStack:stack});
        }
        else{
            throw new Error('Push失败，Page没有注册');
        }
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>返回到前一个页面不带任何返回结果</span></br></br>
     */
    pop(){
        this._popTo({isIgnoreHash:false});
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>返回到前一个页面,附带返回结果和返回行为模式</span></br></br>
     * @param action {String}  页面返回的行为，确认行为值为ok，取消行为值为cancel
     * @param params {JSON} 返回的结果
     */
    popAndReturn(action,params){
        if(action=="ok" || action=="cancel"){
            let pageStack=this.state.pageStack;
            this.popParams={
                action:action,
                params:params,
                pageCode:pageStack[pageStack.length-1].pageCode
            };
            this._popTo({isIgnoreHash:false});
        }
        else{
            throw new Error('页面返回的action参数不正确');
        }
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>返回到某个栈借点，如果不带参数的话，默认当前页出栈</span></br></br>
     * @param name {String} 在传入的页面配置对象Pages中的key值
     */
    popTo(name){
        if(name!=null){
            this._popTo({name:name,isIgnoreHash:false});
        }
        else{
            throw new Error('popTo参数不正确,没有指明要pop到哪里');
        }
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>返回到栈的根节点</span></br></br>
     */
    popToRoot(){
        this._popTo({index:1,isIgnoreHash:false});
    }

    /**
     * @description 页面出栈
     * 包含三种方式：
     * 1.当params.name为空时，执行返回上一页，相当于history.back()
     * 2.当params.name存在，查找栈中页面是否有匹配的，如果匹配则现实该页，栈中该页面后的其他页面全部出栈
     * 3.当params.index存在时，根据index的索引号进行出栈，这里主要用于popToRoot方法，出栈所有页，只保留第一页
     * @param params
     * @private
     */
    _popTo(params){
        var i,
            back,
            index,
            stack=this.state.pageStack,
            slen=stack.length;
        if(slen==1){
            throw new Error("已经是页面栈中的唯一元素，无法再Pop!");
        }
        else if(slen>1){
            //如果指定了要跳转的索引位置
            if(params.index){
                index=params.index;
                back=slen-index;
            }
            //没有名称，默认pop当前页面
            else if(params.name==null){
                index=slen-1;
                back=1;
            }
            else{
                for(i=slen;i--;){
                    if(params.name==stack[i].name){
                        index=i+1;
                        back=slen-i-1;
                        break;
                    }
                }
            }
            //如果找到,进行出栈操作
            if(index!=null){
                stack.splice(index);
                delete stack[index-1].isRestored;
                this.setState({pageStack:stack});
                if(!params.isIgnoreHash){
                    this.stackHash=true;
                    history.go(-back);
                }
            }
            else{
                throw new Error("没有在当前栈找到相应的页面，无法Pop！");
            }
        }
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>替换当前页面，栈的长度不发生变化</span></br></br>
     * @param name {String} 在传入的页面配置对象Pages中的key值
     * @param pageParams {JSON} 页面需要附加的参数，带入Page页面中
     */
    replace(name,pageParams){
        let stack=this.state.pageStack;
        const self=this;
        this._loadPageClass(name,function(){
            const pageId = (new Date()).getTime();
            let hash = name + '/' + pageId;
            if(self.props.pages[name]){
                stack[stack.length-1]={
                    pageId:pageId,
                    name:name,
                    params:pageParams || '',
                    hash:hash
                };
                if(stack.length>1){
                    //history replace
                    self.stackHash=true;
                    self.replaceHash = hash;
                    history.back();
                }
                self.setState({pageStack:stack});
            }
        });
    }

    /**
     * @description
     * </br><span style='font-size:14px;color:#74af1f;'>返回到根节点页面，然后再从根节点页面上push新的页面</span>
     * <ul style='font-size:14px;color:#74af1f;'>
     * <li>1.如果当前就在根节点页面上，则执行push操作</li>
     * <li>2.如果当前name参数为根节点，则执行popToRoot操作 </li>
     * <li>3.如果非以上两种例外，则执行出栈到根节点再压栈的操作</li>
     * <ul>
     * </br>
     * @param name {String} 页面配置对象Pages中的key值
     * @param pageParams {JSON} 页面需要附加的参数，带入Page页面中，参数中不能传入function
     */
    forkTo(name,pageParams){
        let back;
        let stack=this.state.pageStack;
        let slen=stack.length;
        let hash;

        if(slen==1){
            this.push(name,pageParams);
        }
        else if(slen>1){
            //如果name和根页面的name一样，则执行出栈到根节点
            if(stack[0].name==name){
                this.popToRoot();
            }
            else{
                //将栈清除到只剩下根页面
                stack.splice(1);
                back=slen-1;
                //将新的页面压入栈
                const pageId = (new Date()).getTime();
                hash = name+'/'+pageId;
                stack.push({
                    pageId:pageId,
                    name:name,
                    params:pageParams||'',
                    hash:hash
                });
                this.setState({pageStack:stack});
                this.stackHash=true;
                this.replaceHash=hash;
                history.go(-back);
            }
        }
    }

}

export default NavigationController;
